"""
Character sections models
"""

from typing import Optional

from django.db import models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveEntity, EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.helpers import EveEntityIdsMixin
from memberaudit.managers.character_sections_1 import (
    CharacterAssetManager,
    CharacterAttributesManager,
    CharacterContactLabelManager,
    CharacterContactManager,
    CharacterContractBidManager,
    CharacterContractItemManager,
    CharacterContractManager,
)

from ._helpers import AddGenericReprMixin
from .characters import Character
from .constants import CURRENCY_MAX_DECIMALS, CURRENCY_MAX_DIGITS, NAMES_MAX_LENGTH
from .general import Location

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterAsset(AddGenericReprMixin, models.Model):
    """An Eve Online asset belonging to a character."""

    id = models.BigAutoField(primary_key=True)

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="assets"
    )
    item_id = models.PositiveBigIntegerField()

    location = models.ForeignKey(
        Location, on_delete=models.CASCADE, default=None, null=True
    )
    parent = models.ForeignKey(
        "CharacterAsset",
        on_delete=models.CASCADE,
        default=None,
        null=True,
        related_name="children",
    )

    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")
    is_blueprint_copy = models.BooleanField(default=None, null=True, db_index=True)
    is_singleton = models.BooleanField()
    location_flag = models.CharField(max_length=NAMES_MAX_LENGTH)
    name = models.CharField(max_length=NAMES_MAX_LENGTH, default="")
    quantity = models.PositiveIntegerField()

    objects = CharacterAssetManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "item_id"],
                name="functional_pk_characterasset",
            )
        ]

    def __str__(self) -> str:
        return f"{self.character}-{self.item_id}-{self.name_display}"

    @property
    def name_display(self) -> str:
        """Return name of this asset to be displayed to user."""
        name = self.name if self.name else self.eve_type.name
        if self.is_blueprint_copy:
            name += " [BPC]"
        return name

    @property
    def icon_url(self) -> str:
        """Return URL for an icon."""
        variant = self.eve_type.IconVariant.BPC if self.is_blueprint_copy else None
        return self.eve_type.icon_url(variant=variant)

    @property
    def group_display(self) -> str:
        """Return group name of this asset to be displayed to user."""
        return self.eve_type.name if self.name else self.eve_type.eve_group.name


class CharacterAttributes(AddGenericReprMixin, models.Model):
    """The training attributes of a character."""

    character = models.OneToOneField(
        Character,
        primary_key=True,
        on_delete=models.CASCADE,
        related_name="attributes",
        help_text="character these attributes belongs to",
    )

    accrued_remap_cooldown_date = models.DateTimeField(default=None, null=True)
    last_remap_date = models.DateTimeField(default=None, null=True)
    bonus_remaps = models.PositiveIntegerField()
    charisma = models.PositiveIntegerField()
    intelligence = models.PositiveIntegerField()
    memory = models.PositiveIntegerField()
    perception = models.PositiveIntegerField()
    willpower = models.PositiveIntegerField()
    objects = CharacterAttributesManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return str(self.character)


class CharacterCloneInfo(AddGenericReprMixin, models.Model):
    """General clone infos for this character."""

    character = models.OneToOneField(
        Character, on_delete=models.CASCADE, related_name="clone_info"
    )

    home_location = models.ForeignKey(
        Location, on_delete=models.CASCADE, null=True, default=None
    )
    last_clone_jump_date = models.DateTimeField(default=None, null=True)
    last_station_change_date = models.DateTimeField(default=None, null=True)

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return str(self.character)


class CharacterContactLabel(AddGenericReprMixin, models.Model):
    """An Eve Online contact label belonging to a Character."""

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="contact_labels"
    )
    label_id = models.PositiveBigIntegerField()
    name = models.CharField(max_length=NAMES_MAX_LENGTH)

    objects = CharacterContactLabelManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "label_id"], name="functional_pk_characterlabel"
            )
        ]

    def __str__(self) -> str:
        return f"{self.character}-{self.name}"


class CharacterContact(AddGenericReprMixin, models.Model):
    """An Eve Online contact belonging to a Character."""

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="contacts"
    )
    eve_entity = models.ForeignKey(
        EveEntity, on_delete=models.CASCADE, related_name="+"
    )

    is_blocked = models.BooleanField(default=None, null=True)
    is_watched = models.BooleanField(default=None, null=True)
    standing = models.FloatField()
    labels = models.ManyToManyField(CharacterContactLabel, related_name="contacts")

    objects = CharacterContactManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "eve_entity"],
                name="functional_pk_charactercontact",
            )
        ]

    def __str__(self) -> str:
        return f"{self.character}-{self.eve_entity.name}"


class CharacterContract(EveEntityIdsMixin, AddGenericReprMixin, models.Model):
    """An Eve Online contract belonging to a Character"""

    AVAILABILITY_ALLIANCE = "AL"
    AVAILABILITY_CORPORATION = "CO"
    AVAILABILITY_PERSONAL = "PR"
    AVAILABILITY_PUBLIC = "PU"
    AVAILABILITY_CHOICES = (
        (AVAILABILITY_ALLIANCE, _("alliance")),
        (AVAILABILITY_CORPORATION, _("corporation")),
        (AVAILABILITY_PERSONAL, _("private")),
        (AVAILABILITY_PUBLIC, _("public")),
    )
    ESI_AVAILABILITY_MAP = {
        "alliance": AVAILABILITY_ALLIANCE,
        "corporation": AVAILABILITY_CORPORATION,
        "personal": AVAILABILITY_PERSONAL,
        "public": AVAILABILITY_PUBLIC,
    }

    STATUS_OUTSTANDING = "OS"
    STATUS_IN_PROGRESS = "IP"
    STATUS_FINISHED_ISSUER = "FI"
    STATUS_FINISHED_CONTRACTOR = "FC"
    STATUS_FINISHED = "FS"
    STATUS_CANCELED = "CA"
    STATUS_REJECTED = "RJ"
    STATUS_FAILED = "FL"
    STATUS_DELETED = "DL"
    STATUS_REVERSED = "RV"
    STATUS_CHOICES = (
        (STATUS_CANCELED, _("canceled")),
        (STATUS_DELETED, _("deleted")),
        (STATUS_FAILED, _("failed")),
        (STATUS_FINISHED, _("finished")),
        (STATUS_FINISHED_CONTRACTOR, _("finished contractor")),
        (STATUS_FINISHED_ISSUER, _("finished issuer")),
        (STATUS_IN_PROGRESS, _("in progress")),
        (STATUS_OUTSTANDING, _("outstanding")),
        (STATUS_REJECTED, _("rejected")),
        (STATUS_REVERSED, _("reversed")),
    )
    ESI_STATUS_MAP = {
        "canceled": STATUS_CANCELED,
        "deleted": STATUS_DELETED,
        "failed": STATUS_FAILED,
        "finished": STATUS_FINISHED,
        "finished_contractor": STATUS_FINISHED_CONTRACTOR,
        "finished_issuer": STATUS_FINISHED_ISSUER,
        "in_progress": STATUS_IN_PROGRESS,
        "outstanding": STATUS_OUTSTANDING,
        "rejected": STATUS_REJECTED,
        "reversed": STATUS_REVERSED,
    }

    TYPE_AUCTION = "AT"
    TYPE_COURIER = "CR"
    TYPE_ITEM_EXCHANGE = "IE"
    TYPE_LOAN = "LN"
    TYPE_UNKNOWN = "UK"
    TYPE_CHOICES = (
        (TYPE_AUCTION, _("auction")),
        (TYPE_COURIER, _("courier")),
        (TYPE_ITEM_EXCHANGE, _("item exchange")),
        (TYPE_LOAN, _("loan")),
        (TYPE_UNKNOWN, _("unknown")),
    )
    ESI_TYPE_MAP = {
        "auction": TYPE_AUCTION,
        "courier": TYPE_COURIER,
        "item_exchange": TYPE_ITEM_EXCHANGE,
        "loan": TYPE_LOAN,
        "unknown": TYPE_UNKNOWN,
    }

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="contracts"
    )
    contract_id = models.IntegerField()

    acceptor = models.ForeignKey(
        EveEntity,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        related_name="+",
        help_text="Who will accept the contract if character",
    )
    acceptor_corporation = models.ForeignKey(
        EveEntity,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        related_name="+",
        help_text="corporation of acceptor",
    )  # TODO: This field is currently unused
    assignee = models.ForeignKey(
        EveEntity,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        related_name="+",
        help_text="To whom the contract is assigned, can be a corporation or a character",
    )
    availability = models.CharField(
        max_length=2,
        choices=AVAILABILITY_CHOICES,
        help_text="To whom the contract is available",
    )
    buyout = models.DecimalField(
        max_digits=CURRENCY_MAX_DIGITS,
        decimal_places=CURRENCY_MAX_DECIMALS,
        default=None,
        null=True,
    )
    collateral = models.DecimalField(
        max_digits=CURRENCY_MAX_DIGITS,
        decimal_places=CURRENCY_MAX_DECIMALS,
        default=None,
        null=True,
    )
    contract_type = models.CharField(max_length=2, choices=TYPE_CHOICES)
    date_accepted = models.DateTimeField(default=None, null=True)
    date_completed = models.DateTimeField(default=None, null=True)
    date_expired = models.DateTimeField()
    date_issued = models.DateTimeField()
    days_to_complete = models.IntegerField(default=None, null=True)
    end_location = models.ForeignKey(
        Location,
        on_delete=models.CASCADE,
        related_name="contract_end_location",
        default=None,
        null=True,
    )
    for_corporation = models.BooleanField()
    issuer_corporation = models.ForeignKey(
        EveEntity, on_delete=models.CASCADE, related_name="+"
    )
    issuer = models.ForeignKey(EveEntity, on_delete=models.CASCADE, related_name="+")
    price = models.DecimalField(
        max_digits=CURRENCY_MAX_DIGITS,
        decimal_places=CURRENCY_MAX_DECIMALS,
        default=None,
        null=True,
    )
    reward = models.DecimalField(
        max_digits=CURRENCY_MAX_DIGITS,
        decimal_places=CURRENCY_MAX_DECIMALS,
        default=None,
        null=True,
    )
    start_location = models.ForeignKey(
        Location,
        on_delete=models.CASCADE,
        related_name="contract_start_location",
        default=None,
        null=True,
    )
    status = models.CharField(max_length=2, choices=STATUS_CHOICES)
    title = models.CharField(max_length=NAMES_MAX_LENGTH, default="")
    volume = models.FloatField(default=None, null=True)

    objects = CharacterContractManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "contract_id"],
                name="functional_pk_charactercontract",
            )
        ]

    def __str__(self) -> str:
        return f"{self.character}-{self.contract_id}"

    @property
    def is_completed(self) -> bool:
        """Return True when this contract is completed, else False."""
        return self.status in [
            self.STATUS_FINISHED_ISSUER,
            self.STATUS_FINISHED_CONTRACTOR,
            self.STATUS_FINISHED_ISSUER,
            self.STATUS_CANCELED,
            self.STATUS_REJECTED,
            self.STATUS_DELETED,
            self.STATUS_FINISHED,
            self.STATUS_FAILED,
        ]

    @property
    def is_in_progress(self) -> bool:
        """Return True when contract is in progress, else False."""
        return self.status == self.STATUS_IN_PROGRESS

    @property
    def is_failed(self) -> bool:
        """Return True when contract is failed, else False."""
        return self.status == self.STATUS_FAILED

    @property
    def has_expired(self) -> bool:
        """Return True when this contract is expired, else False."""
        return self.date_expired < now()

    @property
    def hours_issued_2_completed(self) -> Optional[float]:
        """Return amount of hours between issues and completed.
        Or return None if still pending.
        """
        if not self.date_completed:
            return None
        deadline = self.date_completed - self.date_issued
        return deadline.days * 24 + (deadline.seconds / 3600)

    def summary(self) -> str:
        """Return summary text for this contract."""
        if self.contract_type == CharacterContract.TYPE_COURIER:
            if not self.start_location or not self.end_location:
                return ""
            return (
                f"{self.start_location.eve_solar_system} >> "
                f"{self.end_location.eve_solar_system} "
                f"({self.volume:.0f} m3)"
            )

        if self.items.filter(is_included=True).count() > 1:
            return _("[Multiple Items]")

        first_item = self.items.first()
        return first_item.eve_type.name if first_item else "(no items)"


class CharacterContractBid(models.Model):
    """A bid belonging to a character contract."""

    contract = models.ForeignKey(
        CharacterContract, on_delete=models.CASCADE, related_name="bids"
    )
    bid_id = models.PositiveIntegerField(db_index=True)

    amount = models.FloatField()
    bidder = models.ForeignKey(EveEntity, on_delete=models.CASCADE, related_name="+")
    date_bid = models.DateTimeField()

    objects = CharacterContractBidManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.contract}-{self.bid_id}"


class CharacterContractItem(AddGenericReprMixin, models.Model):
    """An item belonging to a character contract."""

    contract = models.ForeignKey(
        CharacterContract, on_delete=models.CASCADE, related_name="items"
    )
    record_id = models.PositiveBigIntegerField(db_index=True)

    is_included = models.BooleanField(db_index=True)
    is_singleton = models.BooleanField()
    quantity = models.PositiveIntegerField()
    raw_quantity = models.IntegerField(default=None, null=True)
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")

    objects = CharacterContractItemManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.contract}-{self.record_id}"

    @property
    def is_blueprint_original(self) -> bool:
        """Return True if item is an original blueprint, else False."""
        return self.raw_quantity == -1

    @property
    def is_blueprint_copy(self) -> bool:
        """Return True if item is a blueprint copy, else False."""
        return self.raw_quantity == -2

    @property
    def is_blueprint(self) -> bool:
        """Return True if item is any kind of blueprint, else False."""
        return self.raw_quantity in [-1, -2]

    @property
    def name_display(self) -> str:
        """Return name to be displayed to user."""
        name = self.eve_type.name
        if self.is_blueprint_copy:
            name += " [BPC]"
        return name



"""Django app definition."""

from django.apps import AppConfig

from . import __version__


class MemberAuditConfig(AppConfig):
    name = "memberaudit"
    label = "memberaudit"
    verbose_name = f"Member Audit v{__version__}"

    def ready(self) -> None:
        from . import signals  # noqa: F401 pylint: disable=unused-import



# flake8: noqa
"""
This is a standalone scripts that creates invalid structure ids
in order to test the ESI error rate limit handling for structures
"""

import inspect
import os
import sys

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
myauth_dir = os.path.dirname(os.path.dirname(os.path.dirname(currentdir))) + "/myauth"
sys.path.insert(0, myauth_dir)

import django
from django.apps import apps

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

if not apps.is_installed("memberaudit"):
    raise RuntimeError("The app memberaudit is not installed")

import random

from memberaudit.models import Character, Location

print("Trying to resolve multiple random structure IDs...")
character = Character.objects.first()
token = character.fetch_token()
base_structure_id = 1000000000000

for _ in range(120):
    random_structure_id = random.randint(1000000000000, 1100000000000)
    Location.objects.update_or_create_esi_async(id=random_structure_id, token=token)

print("Completed.")



"""Generic helpers."""

from typing import Any, Optional

import unidecode

from django.contrib.auth.models import User
from django.db import models
from django.utils.text import slugify

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def get_or_create_esi_or_none(
    prop_name: str, dct: dict, model_class: type
) -> Optional[Any]:
    """Get or create a new eveuniverse object from a dictionary entry.

    Return the object on success or None.
    """
    if obj_id := dct.get(prop_name):
        obj = model_class.objects.get_or_create_esi(id=obj_id)[0]  # type: ignore
        return obj

    return None


def get_or_create_or_none(
    prop_name: str, dct: dict, model_class: type
) -> Optional[Any]:
    """Get or creates a Django object from a dictionary entry
    or returns None when the entry is none or does not exist.
    """
    if obj_id := dct.get(prop_name):
        obj = model_class.objects.get_or_create(id=obj_id)[0]  # type: ignore
        return obj

    return None


def get_or_none(prop_name: str, dct: dict, model_class: type) -> Optional[Any]:
    """Get a new Django object from a dictionary entry
    or return None if it does not exist.
    """
    if obj_id := dct.get(prop_name):
        try:
            return model_class.objects.get(id=obj_id)  # type: ignore
        except model_class.DoesNotExist:  # type: ignore
            pass

    return None


def filter_groups_available_to_user(
    groups_qs: models.QuerySet, user: User
) -> models.QuerySet:
    """Filter out groups not available to user, e.g. due to state restrictions."""
    return groups_qs.filter(authgroup__states=None) | groups_qs.filter(
        authgroup__states=user.profile.state
    )


def clear_users_from_group(group):
    """Remove all users from given group.

    Workaround for using the clear method,
    which can create problems due to Auth issue #1268
    """
    # TODO: Refactor once Auth issue is fixed
    for user in group.user_set.all():
        user.groups.remove(group)


def get_unidecoded_slug(app_name: str = "Member Audit") -> str:
    """Get an unidecoded slug from a string.

    :param app_name:
    :type app_name:
    :return:
    :rtype:
    """
    return slugify(unidecode.unidecode(app_name), allow_unicode=True)



from io import StringIO
from unittest.mock import patch

from django.core.management import call_command

from app_utils.testing import TestCase

from memberaudit.tests.testdata.factories import create_character_update_status
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.utils import create_memberaudit_character

MODULE_PATH = "memberaudit.management.commands.memberaudit_stats"


@patch(MODULE_PATH + ".get_input")
class TestStats(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()

    def test_command_should_work_1(self, mock_get_input):
        # given
        out = StringIO()
        character_1001 = create_memberaudit_character(1001)
        character_1002 = create_memberaudit_character(1002)
        create_character_update_status(character_1001)
        create_character_update_status(character_1002)

        cases = ["1", "2", "3", "4"]
        for case in cases:
            mock_get_input.return_value = case

            # when/then
            with self.subTest(input=case):
                call_command("memberaudit_stats", stdout=out)



# Generated by Django 4.0.10 on 2023-10-08 18:12

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0012_add_corporation_roles"),
    ]

    operations = [
        migrations.AlterField(
            model_name="characterrole",
            name="character",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="roles",
                to="memberaudit.character",
            ),
        ),
        migrations.AlterField(
            model_name="characterupdatestatus",
            name="section",
            field=models.CharField(
                choices=[
                    ("assets", "assets"),
                    ("attributes", "attributes"),
                    ("character_details", "character details"),
                    ("contacts", "contacts"),
                    ("contracts", "contracts"),
                    ("corporation_history", "corporation history"),
                    ("fw_stats", "faction warfare statistics"),
                    ("implants", "implants"),
                    ("jump_clones", "jump clones"),
                    ("location", "location"),
                    ("loyalty", "loyalty"),
                    ("mails", "mails"),
                    ("mining_ledger", "mining ledger"),
                    ("online_status", "online status"),
                    ("planets", "planets"),
                    ("roles", "roles"),
                    ("ship", "ship"),
                    ("skills", "skills"),
                    ("skill_queue", "skill queue"),
                    ("skill_sets", "skill sets"),
                    ("standings", "standings"),
                    ("titles", "titles"),
                    ("wallet_balance", "wallet balance"),
                    ("wallet_journal", "wallet journal"),
                    ("wallet_transactions", "wallet transactions"),
                ],
                db_index=True,
                max_length=64,
            ),
        ),
        migrations.CreateModel(
            name="CharacterTitle",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title_id", models.PositiveIntegerField()),
                ("name", models.CharField(max_length=100)),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="titles",
                        to="memberaudit.character",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.AddConstraint(
            model_name="charactertitle",
            constraint=models.UniqueConstraint(
                fields=("character", "title_id"), name="functional_pk_charactertitle"
            ),
        ),
    ]



"""Report views."""

from collections import defaultdict
from typing import Dict, List, Sequence

from django.contrib.auth.decorators import login_required, permission_required
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Count, Exists, OuterRef, Q
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from eveuniverse.core import eveimageserver
from eveuniverse.models import EveType

from allianceauth.authentication.models import get_guest_state_pk
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag
from app_utils.views import bootstrap_icon_plus_name_html, yesno_str

from memberaudit import __title__
from memberaudit.constants import DEFAULT_ICON_SIZE, SKILL_SET_DEFAULT_ICON_TYPE_ID
from memberaudit.models import (
    Character,
    CharacterSkillSetCheck,
    General,
    SkillSet,
    SkillSetGroup,
    SkillSetSkill,
)

from ._common import UNGROUPED_SKILL_SET, add_common_context

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def create_organization_html(eve_character: EveCharacter) -> str:
    """Return character's organization as HTML."""
    return format_html(
        "{}{}",
        eve_character.corporation_name,
        f" [{eve_character.alliance_ticker}]" if eve_character.alliance_name else "",
    )


@login_required
@permission_required("memberaudit.reports_access")
def reports(request) -> HttpResponse:
    """Render view for reports."""
    context = {"page_title": _("Reports")}
    return render(
        request, "memberaudit/reports.html", add_common_context(request, context)
    )


@login_required
@permission_required("memberaudit.reports_access")
def user_compliance_report_data(request) -> JsonResponse:
    """Render data view for user compliance report."""
    users_and_character_counts = (
        General.accessible_users(request.user)
        .exclude(profile__state__pk=get_guest_state_pk())
        .annotate(total_chars=Count("character_ownerships__character", distinct=True))
        .annotate(
            unregistered_chars=Count(
                "character_ownerships",
                filter=Q(character_ownerships__character__memberaudit_character=None),
                distinct=True,
            )
        )
        .select_related(
            "profile__main_character",
            "profile__state",
            "profile__main_character__memberaudit_character",
        )
    )
    user_data = []
    for user in users_and_character_counts:
        if user.profile.main_character:
            main_character = user.profile.main_character
            if user == request.user or request.user.has_perm(
                "memberaudit.characters_access"
            ):
                try:
                    character = main_character.memberaudit_character
                except ObjectDoesNotExist:
                    url = None
                else:
                    url = reverse("memberaudit:character_viewer", args=[character.pk])
            else:
                url = None
            main_name = main_character.character_name
            main_html = bootstrap_icon_plus_name_html(
                main_character.portrait_url(),
                main_character.character_name,
                avatar=True,
                url=url,
            )
            corporation_name = main_character.corporation_name
            organization_html = create_organization_html(main_character)
            alliance_name = (
                main_character.alliance_name if main_character.alliance_name else ""
            )
            is_compliant = user.unregistered_chars == 0
        else:
            main_name = user.username
            main_html = bootstrap_icon_plus_name_html(
                eveimageserver.character_portrait_url(1, size=DEFAULT_ICON_SIZE),
                main_name,
                avatar=True,
                url=url,
            )
            alliance_name = organization_html = corporation_name = ""
            is_compliant = False

        is_registered = user.unregistered_chars < user.total_chars
        user_data.append(
            {
                "id": user.pk,
                "main": {
                    "display": main_html,
                    "sort": main_name,
                },
                "organization": {
                    "display": organization_html,
                    "sort": corporation_name,
                },
                "state": user.profile.state.name,
                "corporation_name": corporation_name,
                "alliance_name": alliance_name,
                "total_chars": user.total_chars,
                "unregistered_chars": user.unregistered_chars,
                "is_registered": is_registered,
                "registered_str": yesno_str(is_registered),
                "is_compliant": is_compliant,
                "compliance_str": yesno_str(is_compliant),
            }
        )
    return JsonResponse({"data": user_data})


@login_required
@permission_required("memberaudit.reports_access")
def corporation_compliance_report_data(request) -> JsonResponse:
    """Render data view for corporation compliance report."""
    relevant_user_ids = list(
        General.accessible_users(request.user)
        .exclude(profile__state__pk=get_guest_state_pk())
        .values_list("id", flat=True)
    )
    corporations = (
        EveCharacter.objects.select_related(
            "userprofile",
            "userprofile__user__character_ownerships__character",
            "userprofile__user__character_ownerships",
        )
        .filter(userprofile__in=relevant_user_ids)
        .values(
            "corporation_id",
            "corporation_name",
            "alliance_id",
            "alliance_name",
            "alliance_ticker",
        )
        .distinct()
        .annotate(mains_count=Count("userprofile", distinct=True))
        .annotate(
            characters_count=Count(
                "userprofile__user__character_ownerships__character", distinct=True
            )
        )
        .annotate(
            unregistered_count=Count(
                "userprofile__user__character_ownerships",
                filter=Q(
                    userprofile__user__character_ownerships__character__memberaudit_character__isnull=True
                ),
                distinct=True,
            )
        )
    )
    data = []
    for corporation in corporations:
        corporation_name = corporation["corporation_name"]
        alliance_ticker = (
            f" [{corporation['alliance_ticker']}]"
            if corporation["alliance_ticker"]
            else ""
        )
        organization_name = f"{corporation_name}{alliance_ticker}"
        alliance_name = (
            corporation["alliance_name"] if corporation["alliance_name"] else ""
        )
        compliance_p = (
            round(
                (corporation["characters_count"] - corporation["unregistered_count"])
                / corporation["characters_count"]
                * 100
            )
            if corporation["characters_count"] > 0
            else 0
        )
        is_compliant = compliance_p == 100
        data.append(
            {
                "id": corporation["corporation_id"],
                "organization_html": {
                    "display": bootstrap_icon_plus_name_html(
                        icon_url=eveimageserver.corporation_logo_url(
                            corporation_id=corporation["corporation_id"],
                            size=DEFAULT_ICON_SIZE,
                        ),
                        name=organization_name,
                    ),
                    "sort": corporation["corporation_name"],
                },
                "corporation_name": corporation["corporation_name"],
                "alliance_name": alliance_name,
                "mains_count": corporation["mains_count"],
                "characters_count": corporation["characters_count"],
                "unregistered_count": corporation["unregistered_count"],
                "compliance_percent": compliance_p,
                "is_compliant": is_compliant,
                "is_partly_compliant": compliance_p >= 85,
                "is_compliant_str": yesno_str(is_compliant),
            }
        )
    return JsonResponse({"data": data})


@login_required
@permission_required("memberaudit.reports_access")
def skill_sets_report_data(request) -> JsonResponse:
    """Render data view for skill sets report."""

    failed_required_skills_qs = SkillSetSkill.objects.filter(
        failed_required_skill_set_checks__pk=OuterRef("pk")
    )
    skill_set_checks_qs = (
        CharacterSkillSetCheck.objects.select_related(
            "character",
            "character__eve_character__character_ownership__user",
            "character__eve_character__character_ownership__user__profile__main_character",
            "character__eve_character__character_ownership__user__profile__state",
            "skill_set",
            "skill_set__ship_type",
        )
        .exclude(
            character__eve_character__character_ownership__user__profile__state__pk=(
                get_guest_state_pk()
            )
        )
        .annotate(has_skills=~Exists(failed_required_skills_qs))
    )
    character_skill_checks: Dict[int, List[CharacterSkillSetCheck]] = defaultdict(list)
    for skill_set_check in skill_set_checks_qs:
        character_skill_checks[skill_set_check.skill_set.pk].append(skill_set_check)

    data = []
    groups_map: Dict[int, str] = SkillSet.objects.compile_groups_map()
    for group_map in groups_map.values():
        group: SkillSetGroup = group_map["group"]
        characters_map = {}
        for skill_set in group_map["skill_sets"]:
            for skill_check in character_skill_checks[skill_set.pk]:
                character: Character = skill_check.character
                if character.pk not in characters_map:
                    characters_map[character.pk] = {
                        "character": character,
                        "character_name": character,
                        "skill_sets": [],
                    }
                if skill_check.has_skills:
                    characters_map[character.pk]["skill_sets"].append(
                        skill_check.skill_set
                    )
        for character_map in characters_map.values():
            data.append(
                _build_skill_set_report_row(
                    group, character_map["character"], character_map["skill_sets"]
                )
            )
    return JsonResponse({"data": data})


def _build_skill_set_report_row(
    group: SkillSetGroup, character: Character, skill_sets: Sequence[SkillSet]
) -> dict:
    if character.main_character:
        main_name = character.main_character.character_name
        main_html = bootstrap_icon_plus_name_html(
            character.main_character.portrait_url(), main_name, avatar=True
        )
        main_corporation = character.main_character.corporation_name
        main_alliance = (
            character.main_character.alliance_name
            if character.main_character.alliance_name
            else ""
        )
        organization_html = format_html(
            "{}{}",
            main_corporation,
            (
                f" [{character.main_character.alliance_ticker}]"
                if character.main_character.alliance_name
                else ""
            ),
        )
    else:
        main_html = main_name = ""
        main_corporation = main_alliance = organization_html = ""

    base_url = reverse("memberaudit:character_viewer", args=[character.pk])
    character_viewer_url = f"{base_url}?tab=skill_sets"
    character_html = bootstrap_icon_plus_name_html(
        character.eve_character.portrait_url(),
        character.eve_character.character_name,
        avatar=True,
        url=character_viewer_url,
    )

    has_required = [
        bootstrap_icon_plus_name_html(
            (
                obj.ship_type.icon_url(
                    DEFAULT_ICON_SIZE, variant=EveType.IconVariant.REGULAR
                )
                if obj.ship_type
                else eveimageserver.type_icon_url(
                    SKILL_SET_DEFAULT_ICON_TYPE_ID, size=DEFAULT_ICON_SIZE
                )
            ),
            obj.name,
        )
        for obj in sorted(skill_sets, key=lambda o: o.name.lower())
    ]
    has_required_html = (
        "<br>".join(has_required)
        if has_required
        else '<i class="fas fa-times boolean-icon-false"></i>'
    )
    group_pk = group.pk if group else 0
    state_name = character.user.profile.state.name if character.user else ""
    return {
        "id": f"{group_pk}_{character.pk}",
        "group": group.name_plus if group else UNGROUPED_SKILL_SET,
        "main": main_name,
        "main_html": main_html,
        "state": state_name,
        "organization_html": organization_html,
        "corporation": main_corporation,
        "alliance": main_alliance,
        "character": character.eve_character.character_name,
        "character_html": character_html,
        "has_required": has_required_html,
        "has_required_str": yesno_str(bool(has_required)),
        "is_doctrine_str": yesno_str(group.is_doctrine if group else False),
        "is_main_str": yesno_str(character.is_main),
    }
